# This file is part of ap_association.
#
# Developed for the LSST Data Management System.
# This product includes software developed by the LSST Project
# (https://www.lsst.org).
# See the COPYRIGHT file at the top-level directory of this distribution
# for details of code ownership.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import numpy as np
import tempfile
import unittest
import yaml

from lsst.ap.association import LoadDiaCatalogsTask, LoadDiaCatalogsConfig
from lsst.dax.apdb import Apdb, ApdbSql, ApdbTables
from lsst.utils import getPackageDir
import lsst.utils.tests
from utils_tests import makeExposure, makeDiaObjects, makeDiaSources, makeDiaForcedSources


def _data_file_name(basename, module_name):
    """Return path name of a data file.

    Parameters
    ----------
    basename : `str`
        Name of the file to add to the path string.
    module_name : `str`
        Name of lsst stack package environment variable.

    Returns
    -------
    data_file_path : `str`
       Full path of the file to load from the "data" directory in a given
       repository.
    """
    return os.path.join(getPackageDir(module_name), "data", basename)


class TestLoadDiaCatalogs(unittest.TestCase):

    def setUp(self):
        # Create an instance of random generator with fixed seed.
        rng = np.random.default_rng(1234)

        self.db_file_fd, self.db_file = tempfile.mkstemp(
            dir=os.path.dirname(__file__))
        self.addCleanup(os.remove, self.db_file)
        self.addCleanup(os.close, self.db_file_fd)

        self.apdbConfig = ApdbSql.init_database(db_url="sqlite:///" + self.db_file)
        self.apdb = Apdb.from_config(self.apdbConfig)

        self.exposure = makeExposure(False, False)

        self.diaObjects = makeDiaObjects(20, self.exposure, rng)
        self.diaSources = makeDiaSources(
            100, self.diaObjects["diaObjectId"].to_numpy(), self.exposure, rng)
        self.diaForcedSources = makeDiaForcedSources(
            200, self.diaObjects["diaObjectId"].to_numpy(), self.exposure, rng)

        self.dateTime = self.exposure.visitInfo.date
        self.apdb.store(self.dateTime.toAstropy(),
                        self.diaObjects,
                        self.diaSources,
                        self.diaForcedSources)

        # These columns are not in the DPDD, yet do appear in DiaSource.yaml.
        # We don't need to check them against the default APDB schema.
        self.ignoreColumns = ["band", "bboxSize", "isDipole", "flags"]

    def testRun(self):
        """Test the full run method for the loader.
        """
        diaLoader = LoadDiaCatalogsTask()
        result = diaLoader.run(self.exposure, self.apdb)

        self.assertEqual(len(result.diaObjects), len(self.diaObjects))
        self.assertEqual(len(result.diaSources), len(self.diaSources))
        self.assertEqual(len(result.diaForcedSources),
                         len(self.diaForcedSources))

    def testLoadDiaObjects(self):
        """Test that the correct number of diaObjects are loaded.
        """
        diaLoader = LoadDiaCatalogsTask()
        region = diaLoader._getRegion(self.exposure)
        diaObjects = diaLoader.loadDiaObjects(region,
                                              self.apdb)
        self.assertEqual(len(diaObjects), len(self.diaObjects))

    def testLoadDiaForcedSources(self):
        """Test that the correct number of diaForcedSources are loaded.
        """
        diaLoader = LoadDiaCatalogsTask()
        region = diaLoader._getRegion(self.exposure)
        diaForcedSources = diaLoader.loadDiaForcedSources(
            self.diaObjects,
            region,
            self.dateTime,
            self.apdb)
        self.assertEqual(len(diaForcedSources), len(self.diaForcedSources))

    def testLoadDiaSources(self):
        """Test that the correct number of diaSources are loaded.

        Also check that they can be properly loaded both by location and
        ``diaObjectId``.
        """
        diaConfig = LoadDiaCatalogsConfig()
        diaLoader = LoadDiaCatalogsTask(config=diaConfig)

        region = diaLoader._getRegion(self.exposure)
        diaSources = diaLoader.loadDiaSources(self.diaObjects,
                                              region,
                                              self.dateTime,
                                              self.apdb)
        self.assertEqual(len(diaSources), len(self.diaSources))

    def test_apdbSchema(self):
        """Test that the default DiaSource schema from dax_apdb agrees with the
        column names defined here in ap_association/data/DiaSource.yaml.
        """
        tableDef = self.apdb.tableDef(ApdbTables.DiaSource)
        apdbSchemaColumns = [column.name for column in tableDef.columns]

        functorFile = _data_file_name("DiaSource.yaml", "ap_association")
        with open(functorFile) as yaml_stream:
            diaSourceFunctor = yaml.safe_load_all(yaml_stream)
            for functor in diaSourceFunctor:
                diaSourceColumns = [column for column in list(functor['funcs'].keys())
                                    if column not in self.ignoreColumns]
            self.assertLess(set(diaSourceColumns), set(apdbSchemaColumns))


class MemoryTester(lsst.utils.tests.MemoryTestCase):
    pass


def setup_module(module):
    lsst.utils.tests.init()


if __name__ == "__main__":
    lsst.utils.tests.init()
    unittest.main()
